typedef unsigned char UCHAR;
typedef unsigned short USHORT;

#include <fcntl.h>
#include <share.h>
#include <io.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "td.h"

// trd prototypes
void trd_onstart();
void trd_ontrack();
void trd_onsector();
void trd_onend();

/*************** teledisk processing section ***************/

TD_IMGHDR  ImgHdr;
TD_INFOHDR InfoHdr;
TD_TRKHDR  TrkHdr;
TD_SECHDR  SecHdr;

// my own error codes
#define ERR_RD		1
#define ERR_CRC		2
#define ERR_FRMT 	3
#define ERR_VER		4
#define ERR_LZ1		5
#define ERR_VOL		6
#define ERR_CORRUPT	7

int TdCompress;
UCHAR szTdFileName[_MAX_PATH];
int hTdFile=-1;
int hTdFile2=-1; // my output in uncompressed form

UCHAR TdSecBuf1[0x4002];
UCHAR TdSecBuf2[0x4000];

int _cdecl td_nextvol(int err)
{
 int hf;
 unsigned dummy;
 TD_IMGHDR imghdr;

 _dos_close(hTdFile);
 hTdFile=-1;
 if(err) return -1;

 szTdFileName[strlen(szTdFileName)-1]++;
 if(_dos_open(szTdFileName, O_RDONLY|SH_DENYWR, &hf)) return -1;
 if(!_dos_read(hf, &imghdr, sizeof(imghdr), &dummy) && dummy==sizeof(imghdr) &&
    imghdr.id[0]==ImgHdr.id[0] && imghdr.id[1]==ImgHdr.id[1] &&
    imghdr.crc==td_crc_count(&imghdr, sizeof(imghdr)-2) &&
    imghdr.vol_id==ImgHdr.vol_id &&
    imghdr.vol==ImgHdr.vol+1) // this increment of expected volume_no is probably missed in original td v2.15 read module (write one does)
   {
    memcpy(&ImgHdr, &imghdr, sizeof(ImgHdr));
    return hTdFile=hf;
   }
 _dos_close(hf);
 return -1;
}

int td_read_raw(void *p, unsigned s)
{
 unsigned dummy;
 int r;
 r=_dos_read(hTdFile, p, s, &dummy);
 if(hTdFile2!=-1 && !r) _dos_write(hTdFile2, p, dummy, &dummy);
 return r;
}

int td_read(void *p, unsigned s)
{
 unsigned dummy, dummy2;
 int r;
 if(!s) return 0;
 if(TdCompress==2)
   {
    r=lz_v2_read(p, s);
    if(hTdFile2!=-1 && !r) _dos_write(hTdFile2, p, s, &dummy2);
   }
 else if(TdCompress==1) return -1;
 else
   {
    for(;;)
      {
       r=_dos_read(hTdFile, p, s, &dummy);
       if(r) return r;
       if(hTdFile2!=-1 && dummy) _dos_write(hTdFile2, p, dummy, &dummy2);
       p=(char*)p+dummy; s-=dummy;
       if(!s) return 0;
       if((hTdFile=td_nextvol(0))==-1) return -1;
      }
   }

 return r;
}

int td_test(void)
{
 int i,j;
 USHORT crc, seclen;

 if(td_read_raw(&ImgHdr, sizeof(ImgHdr))) return ERR_RD;

 if(ImgHdr.id[0]=='T' && ImgHdr.id[1]=='D') TdCompress=0;
 else if(ImgHdr.id[0]=='t' && ImgHdr.id[1]=='d') TdCompress=1;
 else return ERR_FRMT;

 if(ImgHdr.crc!=td_crc_count(&ImgHdr, sizeof(ImgHdr)-2)) return ERR_CRC;

 if(ImgHdr.ver<10 || ImgHdr.ver>21) return ERR_VER;

 if(TdCompress)
   {
    if(ImgHdr.ver>=20) {TdCompress++; lz_v2_init(td_nextvol, hTdFile);}
    else {return ERR_LZ1;}
   }

 if(ImgHdr.vol) return ERR_VOL;

 if(hTdFile2!=-1 && ImgHdr.id[0]=='t' && ImgHdr.id[1]=='d')
   {
    lseek(hTdFile2, 0, SEEK_SET);
    ImgHdr.id[0]='T'; ImgHdr.id[1]='D'; ImgHdr.crc=td_crc_count(&ImgHdr, sizeof(ImgHdr)-2);
    _dos_write(hTdFile2, &ImgHdr, sizeof(ImgHdr), (unsigned*)&i);
    ImgHdr.id[0]='t'; ImgHdr.id[1]='d'; ImgHdr.crc=td_crc_count(&ImgHdr, sizeof(ImgHdr)-2);
   }

 if(ImgHdr.trk_den & 0x80)
   {
    if(td_read(&InfoHdr, 10)) return ERR_RD;
    if(InfoHdr.size>0x248) return ERR_CORRUPT;
    if(td_read(InfoHdr.text, InfoHdr.size)) return ERR_RD;
    if(InfoHdr.crc!=td_crc_count(&InfoHdr.size, InfoHdr.size+8)) return ERR_CRC;
   }

 trd_onstart();

 for(;;)
  {
   if(td_read(&TrkHdr, sizeof(TrkHdr))) return ERR_RD;
   if(TrkHdr.n==0xFF) break;
   if(TrkHdr.crc!=(UCHAR)td_crc_count(&TrkHdr, sizeof(TrkHdr)-1)) return ERR_CRC;

   trd_ontrack();

   for(i=TrkHdr.n; i>0; i--)
     {
      if(td_read(&SecHdr, sizeof(SecHdr))) return ERR_RD;
      crc=(UCHAR)td_crc_count(&SecHdr, sizeof(SecHdr)-1);
      if((!(SecHdr.n&0xF8)) == (!(SecHdr.flag&0x30)))
        {
         if(td_read(&seclen, 2)) return ERR_RD;
         if(seclen>0x4000) return ERR_CORRUPT;
         if(td_read(TdSecBuf1+2, seclen)) return ERR_RD;
         TdSecBuf1[0]=seclen; TdSecBuf1[1]=seclen>>8;
         if((seclen=td_unpack_trk_data(TdSecBuf2, TdSecBuf1, &j))==0xFFFF) return ERR_CORRUPT;
         crc=(UCHAR)td_crc_count(TdSecBuf2, seclen);
         // in original td code was (ptr, len, crc), but third arg wasn't used anyway.
         // de facto, either data or hdr is crc unprotected, he-he ;)
        }
      if(SecHdr.crc!=(UCHAR)crc) return ERR_CRC;

      trd_onsector();
     }
  }

 trd_onend();

 return 0;
}
/***********************************************************/

/***************** trd processing section *****************/
int trd_hfile=-1;
int trd_track, trd_valid;
USHORT trd_track_mask;
UCHAR trd_buf[0x1400];

void trd_write(USHORT s)
{
 unsigned dummy;
 if(trd_hfile!=-1) _dos_write(trd_hfile, trd_buf, s, &dummy);
}

void trd_onstart()
{
 trd_track=-1;
 trd_valid=1;
}

void trd_ontrack()
{
 if(trd_valid && trd_track>=0)
   {
    if(trd_valid==2 && trd_track_mask==0)
      {
       trd_write(16*0x100);
      }
    else if(trd_valid==3 && (trd_track_mask&0x010F)==0)
      {
       trd_write(5*0x400);
      }
    else trd_valid=0;
   }
 trd_track++;
 if((TrkHdr.c != (trd_track>>1)) || (TrkHdr.h != (trd_track&1))) trd_valid=0;
 trd_track_mask=0xFFFF;
}

void trd_onsector()
{
 USHORT mask;

 if(!trd_valid) return;
 if((SecHdr.c==(trd_track>>1)) && ((SecHdr.flag&0x76)==0))
   {
    if((SecHdr.n==1 && SecHdr.r>=1 && SecHdr.r<=16) && (trd_valid==1 || trd_valid==2))
      {
       mask=1<<(SecHdr.r-1);
       if(trd_track_mask & mask)
         {
          trd_track_mask &= ~mask;
          memcpy(trd_buf+0x100*(SecHdr.r-1), TdSecBuf2, 0x100);
         }
       trd_valid=2;
      }
    else if(SecHdr.n==3 && (SecHdr.r==9 || (SecHdr.r>=1 && SecHdr.r<=4)) && (trd_valid==1 || trd_valid==3))
      {
       mask=1<<(SecHdr.r-1);
       if(trd_track_mask & mask)
         {
          trd_track_mask &= ~mask;
          if(SecHdr.r==9) memcpy(trd_buf+0x1000, TdSecBuf2, 0x400);
          else memcpy(trd_buf+0x400*(SecHdr.r-1), TdSecBuf2, 0x400);
         }
       trd_valid=3;
      }
   }
}

void trd_onend()
{
 if(++trd_track!=160)
   {
    if(trd_track>160 && trd_valid==2) trd_valid=4;
    else trd_valid=0;
   }
 if(trd_valid)
   {
    if(((trd_valid==2)||(trd_valid==4)) && trd_track_mask==0)
      {
       trd_write(16*0x100);
      }
    else if(trd_valid==3 && (trd_track_mask&0x010F)==0)
      {
       trd_write(5*0x400);
      }
    else trd_valid=0;
   }
 if(trd_valid) trd_valid+=10;
}
/**********************************************************/

void usage()
{
 printf("\nUsage:  tele_cvt tdfile [-td0 tdfile2] [-trd trdfile]\n"
        "where:\n"
        "    tdfile   input file in teledisk format\n"
        "    tdfile2  optional output file in teledisk-normal format\n"
        "    trdfile  optional output file in trd format\n"
        "(just a test of input if no output(s) specified)\n");
 exit(1);
}

UCHAR LZbuf[_LZ_V2_DATABUFFER_SIZE];

int main(int argc, char **argv)
{
 UCHAR *p, *ptrd, *ptd0;
 int err;

 printf("Teledisk file converter v1.1  07/09/98  MK\n");
 if(argc<2) usage();

 ptrd=ptd0=0; szTdFileName[0]=0;

 while(--argc)
   {
    if(*(p=*++argv)=='-' || *p=='/')
      {
       if(!strnicmp(p+1,"trd",3))
         {
          if(p[4]) ptrd=p+4;
          else if(--argc) ptrd=*++argv;
          else usage();
         }
       else if(!strnicmp(p+1,"td0",3))
         {
          if(p[4]) ptd0=p+4;
          else if(--argc) ptd0=*++argv;
          else usage();
         }
       else usage();
      }
     else
      {
       if(!szTdFileName[0]) strcpy(szTdFileName, p);
       else usage();
      }
   }

 if(!szTdFileName[0]) usage();

 lz_v2_databuffer=LZbuf;

 if(_dos_open(szTdFileName, O_RDONLY|SH_DENYWR, &hTdFile))
   {printf("Can't open %s!\n", szTdFileName); return -2;}

 if(ptrd && _dos_creat(ptrd, _A_NORMAL, &trd_hfile))
   {printf("Can't create %s!\n", ptrd); _dos_close(hTdFile); return -2;}

 if(ptd0 && _dos_creat(ptd0, _A_NORMAL, &hTdFile2))
   {printf("Can't create %s!\n", ptrd); _dos_close(hTdFile);
    if(ptrd) {_dos_close(trd_hfile); unlink(ptrd);}
    return -2;}

 err=td_test();

 switch(err)
   {
    case 0: printf("No Fatal error(s)\n"); p=0; break;
    case ERR_FRMT: p="Non-teledisk file"; break;
    case ERR_RD: p="Unexpected end of file"; break;
    case ERR_CORRUPT: p="Some internal structures are too large"; break;
    case ERR_CRC: p="CRC failed"; break;
    case ERR_VER: p="Incorrect teledisk version"; break;
    case ERR_LZ1: p="No support for compression v1.x"; break;
    case ERR_VOL: p="Not a main volume"; break;
   }
 if(p) printf("Error: %s\n", p);

 if(ptrd && !err)
   {
    if(trd_valid==14) sprintf(trd_buf, " %d%s tracks", trd_track>>1, (trd_track&1)?".5":"");
     else trd_buf[0]=0;
    if(trd_valid>10) printf("Trd file created successfully (type=%sdos%s)\n", (trd_valid==13)?"is":"tr", trd_buf);
     else printf("Trd file creation error (incorrect input format?)\n");
   }

 if(hTdFile>0) _dos_close(hTdFile);
 if(ptrd) {_dos_close(trd_hfile); if(err || trd_valid<10) unlink(ptrd);}
 if(ptd0) {_dos_close(hTdFile2); if(err) unlink(ptd0);}

 return 0;
}
